#!/usr/bin/env python


# $Id: rst2html.py 4564 2006-05-21 20:44:42Z wiemann $
# Author: David Goodger <goodger@python.org>
# Copyright: This module has been placed in the public domain.

"""
A front end to the Docutils Publisher, registering the GoDiagramDirective 
and producing HTML.  Based on rst2html.py by David Goodger, in the public 
domain.
"""

try:
    import locale
    locale.setlocale(locale.LC_ALL, '')
except:
    pass

from docutils.core import publish_cmdline, default_description

from rstgo.rst import setup_pelican

setup_pelican()

description = ('Generates (X)HTML documents from standalone reStructuredText '
               'sources.  ' + default_description)

publish_cmdline(writer_name='html', description=description)




class MalformedGoDiagram(ValueError): 
    pass

class GoDiagramParser(object):
    def __init__(self):
        self.line_prefix = ''
        self.top_edge = False
        self.left_edge = False
        self.right_edge = False
        self.bottom_edge = False
        self.board_width = 0
        self.board_height = 0
        self.initial_black = set()
        self.initial_white = set()
        self.sequence_moves = {}
        self.annotated_moves = {}
        self.board_parsed = False

    def board_size(self):
        return self.board_width, self.board_height

    def parse(self, board):
        board = iter(board)
        self.parse_header(board.next())
        for line in board:
            if not self.board_parsed:
                self.parse_board(line)
            else: 
                self.parse_footer(line)

    def parse_header(self, line):
        line = line.strip()
        if line.startswith('$$'):
            self.line_prefix = '$$'
        if '---' in line:
            line = line[len(self.line_prefix):]
            line = line.strip()
            # top border:
            if any(c != '-' for c in line):
                raise MalformedGoDiagram('Top edge must consist of only "-" characters')
            self.top_edge=True
        else:
            self.parse_board(line)
                
    def parse_board(self, line):
        line = line.strip()
        if not line.startswith(self.line_prefix):
            raise MalformedGoDiagram("All lines must start with $$ if any lines start with $$")
        line = line[len(self.line_prefix):]
        line = line.strip()
        if '---' in line:
            self.board_parsed = True
            self.parse_footer(line)
        else:
            self.board_height += 1
            if line[0] == '|':
                self.left_edge = True
                line = line[1:]
            if line[-1] == '|':
                self.right_edge = True
                line = line[:-1]
            points = line.split()
            if not self.board_width:
                self.board_width = len(points)
            elif self.board_width != len(points):
                raise MalformedGoDiagram("All lines must be the same length")
            for i, point in enumerate(points):
                coord = (i, self.board_height - 1)
                if point == 'X':
                    self.initial_black.add(coord)
                elif point == 'O':
                    self.initial_white.add(coord)
                elif point in list('1234567890'):
                    self.sequence_moves[point] = coord
                elif point in list('abcdefghijklmnopqrstuvwxyz'):
                    self.annotated_moves[point] = coord
                elif point in list('.,'):
                    #Empty point
                    pass
                else:
                    raise MalformedGoDiagram('Unrecognized board marking: "%s"' % point)
                    
    def parse_footer(self, line):
        if '---' in line:
            self.bottom_edge = True
            if any(c != '-' for c in line):
                raise MalformedGoDiagram('Top edge must consist of only "-" characters')
        else:
            #Not handling other footer information yet.
            pass




from itertools import cycle
import os

from PIL import Image


class TileSet(object):

    def __init__(self, name='tiny'):
        self.name = name
        self.tile_size = (14, 14)
        self.resource_dir = '%s/resources/%s' % (os.path.dirname(__file__), name)
           
        self.tile_cache = {}
        self.valid_tiles = [
            'tl-corner', 'tr-corner', 'bl-corner', 'br-corner',
            't-side', 'b-side', 'l-side', 'r-side',
            'point', 'starpoint', 'empty',
            'black', 'white',
        ]

    def __getitem__(self, tile):
        if tile not in self.valid_tiles:
            raise KeyError
        if tile not in self.tile_cache:
            self.tile_cache[tile] = Image.open('%s/%s.png' % (self.resource_dir, tile))
        return self.tile_cache[tile]

    def annotate(self, tile, annotation):
        translations = {
            'triangle': u'\u25be',
            'square': u'\u25aa',
            'bullet': u'\u2022',
        }
        annotation = translations.get(annotation, str(annotation))
        annotated_tile_name = '-'.join([tile, annotation])
        if annotated_tile_name not in self.tile_cache:
            tile_object = self[tile].copy()
            self.tile_cache[annotated_tile_name] = tile_object
        return self.tile_cache[annotated_tile_name]
      
    def render(self, diagram):
        image = Image.new('RGBA', 
            (diagram.board.size[0] * self.tile_size[0], 
             diagram.board.size[1] * self.tile_size[1])
        )
        for x in xrange(diagram.board.size[0]):
            for y in xrange(diagram.board.size[1]):
        	point = x, y
                if point in diagram.initial_black:
                    tile = 'black'
                elif point in diagram.initial_white:
                    tile = 'white'
                
                elif diagram.board.left_edge and x == 0:
                    if diagram.board.top_edge and y == 0:
                       tile = 'tl-corner'
                    elif diagram.board.bottom_edge and y == diagram.board.size[1] - 1:
                       tile = 'bl-corner'
                    else:
                       tile = 'l-side'
                elif diagram.board.right_edge and x == diagram.board.size[0] - 1:
                    if diagram.board.top_edge and y == 0:
                       tile = 'tr-corner'
                    elif diagram.board.bottom_edge and y == diagram.board.size[1] - 1:
                       tile = 'br-corner'
                    else:
                       tile = 'r-side'
                elif diagram.board.top_edge and y == 0:
                    tile='t-side'
                elif diagram.board.bottom_edge and y == diagram.board.size[1] - 1:
                    tile='b-side'
                elif (x, y) in diagram.board.star_points:
                    tile = 'starpoint'
                else:
                    tile = 'point'
                box = (x*self.tile_size[0], y*self.tile_size[1],
                       (x+1)*self.tile_size[0], (y+1)*self.tile_size[1])
                image.paste(self[tile], box)

        for i in diagram.sequence:
            if i % 2:
                player = diagram.first
            else:
        	player = diagram.second
            x, y = move
            box = (x*self.tile_size[0], y*self.tile_size[1],
                   (x+1)*self.tile_size[0], (y+1)*self.tile_size[1])
            tile = self.annotate(self[player], i)
            image.paste(tile, box)
        return image


class Board(object):

    def __init__(self, size=(19, 19), left_edge=True, right_edge=True, top_edge=True, bottom_edge=True):
        self.size=size
        self.top_edge = top_edge
        self.left_edge = left_edge
        self.right_edge = right_edge
        self.bottom_edge = bottom_edge
        self.set_star_points() 
        
    def set_star_points(self):
        x_ranks = set()
        y_ranks = set()
        if self.left_edge and not self.right_edge:
            if self.size[0] >= 4:
                x_ranks.add(3)
            if self.size[0] >= 10:
                x_ranks.add(9)
            if self.size[0] >= 16:
                x_ranks.add(15)

        elif self.right_edge and not self.left_edge:
            if self.size[0] >= 4:
                x_ranks.add(self.size[0] - 4)
            if self.size[0] >= 10:
                x_ranks.add(self.size[0] - 10)
            if self.size[0] >= 16:
                x_ranks.add(self.size[0] - 16)
            
        elif self.left_edge and self.right_edge:
            if self.size[0] >=9:
                x_ranks.add(3)
                x_ranks.add(self.size[0] - 4)
            if self.size[0] % 2 and self.size[0] >= 15:
                x_ranks.add((self.size[0] - 1) // 2)

        if self.top_edge and not self.bottom_edge:
            if self.size[1] >= 4:
                y_ranks.add(3)
            if self.size[1] >= 10:
                y_ranks.add(9)
            if self.size[1] >= 16:
                y_ranks.add(15)
 
        elif not self.top_edge and self.bottom_edge:
            if self.size[1] >= 4:
                y_ranks.add(self.size[1] - 4)
            if self.size[1] >= 10:
                y_ranks.add(self.size[1] - 10)
            if self.size[1] >= 16:
                y_ranks.add(self.size[1] - 16)

        elif self.top_edge and self.bottom_edge:
            if self.size[1] >=9:
                y_ranks.add(3)
                y_ranks.add(self.size[1] - 4)
            if self.size[1] % 2 and self.size[1] >= 15:
                y_ranks.add((self.size[1] - 1) // 2)
    
        self.star_points = set()
        for x in x_ranks:
            for y in y_ranks:
                self.star_points.add((x, y))


class GoDiagram(object):

    @classmethod
    def load_diagram_from_parser(cls, parser):
        board = Board(
            size=parser.board_size(),
            top_edge=parser.top_edge,
            left_edge=parser.left_edge,
            right_edge=parser.right_edge,
            bottom_edge=parser.bottom_edge,
        )
        return cls(board=board,
            initial_black = parser.initial_black,
            initial_white = parser.initial_white,
            sequence = parser.sequence_moves,
        )

    def __init__(self, board=Board(), initial_black=None, initial_white=None, sequence=None, first='black'):
        self.board = board
        self.initial_black = initial_black or set()
        self.initial_white = initial_white or set()
        self.sequence = sequence or {}
        self.first = first
        if self.first == 'black':
            self.second = 'white'
        elif self.first == 'white':
            self.second = 'black'
        else:
            self.first = 'black'
            self.second = 'white'
        self.tiles = TileSet('tiny')

    def render(self):
        self.image = self.tiles.render(self)

    def show(self):
        self.image.show()

    def save(self, filename, format):
        self.image.save(filename, format)
        
if __name__ == '__main__':
    black_set = set([(4, 4), (4, 5)])
    white_set = set([(3, 2), (7, 3), (8, 3)])
    sequence = {1: (2, 2), 2: (1,2), 3: (2,3), 4: (3,1)}
    
    board = Board(
        (9, 9), 
    )
    Diagram(board=board, 
            initial_black=black_set, 
            initial_white=white_set, 
            sequence=sequence).draw()



from __future__ import division, absolute_import
import os.path
import posixpath
import tempfile

from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives.images import Image as ImageDirective

from . import parsers, board

class godiagram_node(nodes.General, nodes.Element):
    pass

class GoDiagramDirective(ImageDirective):

    """Renders a go diagram.  This will take an ascii go diagram and return
    an image node which points to the new image
   
    Looks like:
  
    ..go:: filename.png

        $$ -----------
        $$ | . . . . .
        $$ | . . O . .
        $$ | . X O . .
        $$ | . X X . .
        $$ | . . . . .

    It will only render png images.
    """
    
    has_content = True
    required_arguments = 0
    optional_arguments = 1
    option_spec = {
        'alt': directives.unchanged,
    }

    def run(self):
        """Do it."""

        self.assert_has_content()
        node = godiagram_node()
        node['content'] = self.content
        node['alt'] = self.options.get('alt', '')
        if self.arguments:
            node['filename'] = self.arguments[0] 
        else:
            node['filename'] = None
        return [node]

def render_godiagram(self, node, imgbase='.', outbase='.'):

    outfile = node['filename']
    if outfile and outbase:
        imgpath = posixpath.join(imgbase, outfile)
        outpath = os.path.join(outbase, outfile)
        outdir = os.path.dirname(outpath)
        if not os.path.exists(outdir):
            os.makedirs(outdir)
    else:
        imgpath = None
        outpath = tempfile.TemporaryFile()

    parser = parsers.GoDiagramParser()
    parser.parse(node['content'])
    diagram = board.GoDiagram.load_diagram_from_parser(parser)
    diagram.render()
    diagram.save(outpath, 'PNG')
    return imgpath, outpath 

def render_godiagram_html(self, node, imgbase, outbase):
    imgpath, outpath = render_godiagram(self, node, imgbase, outbase)
    self.body.append(self.starttag(node, 'p', CLASS='godiagram'))
  
    if not imgpath:
        try:
            outpath.seek(0)
            imgpath = 'data:image/png;base64,%s' % outpath.read().encode('base64').replace('\n', '')
        finally:
            outpath.close()
    self.body.append('<img src="%s" alt="%s"/>\n' % (imgpath, node['alt']))
    self.body.append('</p>')


# HTML Visitor for godiagram_nodes
def html_visit_godiagram_node(self, node):
    if hasattr(self, 'builder'):
        # Sphinx
        imgbase = getattr(self.builder, 'imgpath')
        outbase = os.path.join(getattr(self.builder, 'outdir'), '_images')
    elif False:
        # Figure out how to do this for Pelican
        pass
    else:
        # Otherwise, fall back to 
        imgbase = None
        outbase = None
    render_godiagram_html(self, node, imgbase, outbase)

def setup(app):
    """setup method to allow sphinx to render go diagrams"""
    app.add_node(godiagram_node,
        html=(html_visit_godiagram_node, lambda self, node: None),
    )
    app.add_directive('go', GoDiagramDirective)

def setup_pelican():
    directives.register_directive('go', GoDiagramDirective)
    from docutils.writers.html4css1 import HTMLTranslator
    setattr(HTMLTranslator, 'visit_godiagram_node', html_visit_godiagram_node)
    setattr(HTMLTranslator, 'depart_godiagram_node', lambda self, node: None)




from unittest2 import TestCase

from rstgo.board import Board
from rstgo.parsers import GoDiagramParser

class ParserTestCase(TestCase):

    def test_parsing_full_empty_board(self):
        board = [
            '  $$  ------------------------------------- \n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$ |. . . . . . . . . . . . . . . . . . .|\n',
            '  $$  ------------------------------------- \n',
        ]
        parser = GoDiagramParser()
        parser.parse(board)
        self.assertEqual(parser.board_size(), (19, 19))
        self.assertEqual(parser.initial_black, set())
        self.assertEqual(parser.initial_white, set())
        self.assertEqual(parser.top_edge, True)
        self.assertEqual(parser.left_edge, True)
        self.assertEqual(parser.right_edge, True)
        self.assertEqual(parser.bottom_edge, True)
 
    def test_parsing_empty_small_full_board(self):
        board = [
            '  $$  ----------------- \n',
            '  $$ |. . . . . . . . .|\n',
            '  $$ |. . . . . . . . .|\n',
            '  $$ |. . . . . . . . .|\n',
            '  $$ |. . . . . . . . .|\n',
            '  $$ |. . . . . . . . .|\n',
            '  $$ |. . . . . . . . .|\n',
            '  $$ |. . . . . . . . .|\n',
            '  $$ |. . . . . . . . .|\n',
            '  $$ |. . . . . . . . .|\n',
            '  $$  ----------------- \n',
        ]
        parser = GoDiagramParser()
        parser.parse(board)
        self.assertEqual(parser.board_size(), (9, 9))
        self.assertEqual(parser.initial_black, set())
        self.assertEqual(parser.initial_white, set())
        self.assertEqual(parser.top_edge, True)
        self.assertEqual(parser.left_edge, True)
        self.assertEqual(parser.right_edge, True)
        self.assertEqual(parser.bottom_edge, True)

    def test_parsing_partial_board(self):
        board = [
            '  $$ ----------------- \n',
            '  $$ . . . . . . . . .|\n',
            '  $$ . . . . . . . . .|\n',
            '  $$ . . . . . . . . .|\n',
            '  $$ . . . . . . . . .|\n',
            '  $$ . . . . . . . . .|\n',
            '  $$ . . . . . . . . .|\n',
            '  $$ . . . . . . . . .|\n',
            '  $$ . . . . . . . . .|\n',
            '  $$ . . . . . . . . .|\n',
        ]
        parser = GoDiagramParser()
        parser.parse(board)
        self.assertEqual(parser.top_edge, True)
        self.assertEqual(parser.left_edge, False)
        self.assertEqual(parser.right_edge, True)
        self.assertEqual(parser.bottom_edge, False)

    def test_parsing_partial_board_with_stones(self):
        board = [
            '  $$  ----------------- \n',
            '  $$ |. . . . . . . . . \n',
            '  $$ |. . . . . . . . . \n',
            '  $$ |. X . . . . . . . \n',
            '  $$ |. O . . . . . . . \n',
            '  $$ |. . . . . . . . . \n',
            '  $$ |. . . . . . . . . \n',
            '  $$ |. . . . . . . . . \n',
            '  $$ |. . . . . . . . . \n',
            '  $$ |. . . . . . . . . \n',
        ]
        parser = GoDiagramParser()
        parser.parse(board)
        self.assertEqual(len(parser.initial_black), 1)
        self.assertIn((1, 2), parser.initial_black)
        self.assertIn((1, 3), parser.initial_white)


class StarpointTestCase(TestCase):
    def test_standard_star_points(self):
        board = Board()
        standard_points=set([(3, 3), (3, 9), (3, 15),
                      (9, 3), (9, 9), (9, 15),
                      (15, 3), (15, 9), (15, 15)])
        self.assertEqual(standard_points, board.star_points)

    def test_no_middle_points_on_even_sizes(self):
        board = Board((18, 19))
        self.assertEqual(len(board.star_points), 6)
        board = Board((18, 18))
        self.assertEqual(len(board.star_points), 4)

    def test_no_middle_points_below_fifteen_rows(self):
        board = Board((15, 15))
        self.assertEqual(len(board.star_points), 9)
        board = Board((13, 13))
        self.assertEqual(len(board.star_points), 4)

    def test_no_points_below_nine_rows(self):
        board = Board((9, 9))
        self.assertEqual(len(board.star_points), 4)
        board = Board((8, 8))
        self.assertEqual(len(board.star_points), 0)

class EdgeTestCase(TestCase):
    pass
